// Shapefile parser, following the specification at
// http://www.esri.com/library/whitepapers/pdfs/shapefile.pdf

SHP = {
    NULL: 0,
    POINT: 1,
    POLYLINE: 3,
    POLYGON: 5
};


SHPParser = function () {
};


SHPParser.prototype.parse = function (arrayBuffer) {
    var o = {},
        dv = new DataView(arrayBuffer),
        idx = 0;
    o.fileCode = dv.getInt32(idx, false);
    if (o.fileCode !== 0x0000270a) {
        throw (new Error("Unknown file code: " + o.fileCode));
    }
    idx += 6 * 4;
    o.wordLength = dv.getInt32(idx, false);
    o.byteLength = o.wordLength * 2;
    idx += 4;
    o.version = dv.getInt32(idx, true);
    idx += 4;
    o.shapeType = dv.getInt32(idx, true);
    idx += 4;
    o.minX = dv.getFloat64(idx, true);
    o.minY = dv.getFloat64(idx + 8, true);
    o.maxX = dv.getFloat64(idx + 16, true);
    o.maxY = dv.getFloat64(idx + 24, true);
    o.minZ = dv.getFloat64(idx + 32, true);
    o.maxZ = dv.getFloat64(idx + 40, true);
    o.minM = dv.getFloat64(idx + 48, true);
    o.maxM = dv.getFloat64(idx + 56, true);
    idx += 8 * 8;
    o.records = [];
    while (idx < o.byteLength) {
        var record = {};
        record.number = dv.getInt32(idx, false);
        idx += 4;
        record.length = dv.getInt32(idx, false);
        idx += 4;
        try {
            record.shape = this.parseShape(dv, idx);
        } catch (e) {
            console.log(e, record);
        }
        idx += record.length * 2;
        o.records.push(record);
    }
    return o;
}


SHPParser.prototype.parseShape = function (dv, idx) {
    var i = 0,
        c = null,
        shape = {};
    shape.type = dv.getInt32(idx, true);
    idx += 4;
    switch (shape.type) {
        case SHP.NULL: // Null
            break;
        case SHP.POINT: // Point (x,y)
            shape.content = {
                x: dv.getFloat64(idx, true),
                y: dv.getFloat64(idx + 8, true)
            };
            break;
        case 13: // PolylineZ
        case SHP.POLYLINE: // Polyline (MBR, partCount, pointCount, parts, points)
        case SHP.POLYGON: // Polygon || Polyline (MBR, partCount, pointCount, parts, points)
            c = shape.content = {
                minX: dv.getFloat64(idx, true),
                minY: dv.getFloat64(idx + 8, true),
                maxX: dv.getFloat64(idx + 16, true),
                maxY: dv.getFloat64(idx + 24, true),
                numParts: dv.getInt32(idx + 32, true),
                numPoints: dv.getInt32(idx + 36, true) * 2
            };
            c.parts = new Int32Array(c.numParts);
            c.points = new Float64Array(c.numPoints);
            idx += 40;
            for (i = 0; i < c.parts.length; i++) {
                c.parts[i] = dv.getInt32(idx, true);
                idx += 4;
            }
            for (i = 0; i < c.points.length; i++) {
                c.points[i] = dv.getFloat64(idx, true);
                idx += 8;
            }
            if (shape.type === 13) {
                idx += 48 + c.numParts * 16
            }
            break;
        case 8: // MultiPoint (MBR, pointCount, points)
        case 11: // PointZ (X, Y, Z, M)
            shape.content = {
                x: dv.getFloat64(idx, true),
                y: dv.getFloat64(idx + 8, true),
                z: dv.getFloat64(idx + 16, true),
                m: dv.getFloat64(idx + 24, true)
            }
            break;
        case 15: // PolygonZ
        case 18: // MultiPointZ
        case 21: // PointM (X, Y, M)
        case 23: // PolylineM
        case 25: // PolygonM
        case 28: // MultiPointM
        case 31: // MultiPatch
            throw new Error("Shape type not supported: " + shape.type)
        default:
            throw new Error("Unknown shape type at " + (idx - 4) + ': ' + shape.type);
    }
    return shape;
};
