/**
 * @author ADS
 * 高斯正反算实现
 */

/**
 * 椭球体
 * @param {string} code 编号
 * @param options
 * @param {number} options.a 长半轴
 * @param {number} options.b 短半轴
 * @param {number} [options.f] 扁率
 * @param {number} [options.e] 第一偏心率
 * @param {number} [options.e2] 第一偏心率平方
 * @param {number} [options.e'] 第二偏心率
 * @param {number} [options.e'2] 第二偏心率平方
 * @param {number} [options.x0=0] x轴偏移
 * @constructor
 */
const Ellipsoid = function (code, options) {
    // const {sin, sqrt, abs, cos, tan, PI} = Math;
    const sin = Math.sin, sqrt = Math.sqrt, abs = Math.abs, cos = Math.cos, tan = Math.tan, PI = Math.PI
    let a = options.a, b = options.b, f = options.f, e = options.e, e_2 = options["e'"], x0 = options.x0 || 0;
    if (!a || a < 0 || isNaN(a)) {
        throw Error('major semi axis must be a number greater than zero')
    }
    if (!b || b < 0 || isNaN(b)) {
        throw Error('minor semi-axis must be a number greater than zero')
    }
    if (!f) {
        f = (a - b) / a;
    }
    if (!e) {
        e = sqrt(a * a - b * b) / a;
    }
    if (!e_2) {
        e_2 = sqrt(a * a - b * b) / b;
    }

    this.code = code;
    this.a = a;
    this.b = b;
    this.f = f;
    this.e = e;
    this["e'"] = e_2;
    this.x0 = x0;
    let e2 = options.e2 || e * e;
    let e_22 = options["e'2"] || e_2 * e_2;

    let m0 = a * (1 - e2);
    let m2 = (3 * e2 * m0) / 2;
    let m4 = (5 * e2 * m2) / 4;
    let m6 = (7 * e2 * m4) / 6;
    let m8 = (9 * e2 * m6) / 8;
    let a0 = m0 + m2 / 2 + 3 * m4 / 8 + 5 * m6 / 16 + 35 * m8 / 128;
    let a2 = m2 / 2 + m4 / 2 + 15 * m6 / 32 + 7 * m8 / 16;
    let a4 = m4 / 8 + 3 * m6 / 16 + 7 * m8 / 16;
    let a6 = m6 / 32 + m8 / 16;
    let a8 = m8 / 128;

    function f1(lng0) {
        return -(a2 * sin(2 * lng0)) / 2 + (a4 * sin(4 * lng0)) / 4 - (a6 * sin(6 * lng0)) / 6 + (a8 * sin(8 * lng0)) / 8;
    }

    // 迭代法反算子午线弧长
    function inverseMeridianArc(x) {
        let lng0 = x / a0;
        while (true) {
            const lng = (x - f1(lng0)) / a0;
            if (abs(lng0 - lng) < 10e-16) {
                break;
            }
            lng0 = lng;
        }
        return lng0
    }

    function correctMeridianArc(lng) {
        return a0 * lng + f1(lng)
    }

    this.correctGauss = function (lng, lat) {

    }

    this.inverseGauss = function (x, y, l0) {
        if (typeof l0 !== 'number' && y / 1000000 >= 1) {
            const bandNum = Math.floor(y / 1000000);
            l0 = 3 * bandNum;
            y = y % 1000000;
        }
        y -= x0;
        // 底点维度
        const bf = inverseMeridianArc(x);
        const sbf = sin(bf), sbf2 = sbf * sbf, cbf = cos(bf), cbf2 = cbf * cbf, tf = tan(bf);
        const wf = sqrt(1 - e2 * sbf2);
        const nf = a / wf, mf = a * (1 - e2) / (wf * wf * wf), qf = e_22 * cbf2;
        const y2 = y * y, y3 = y2 * y, y4 = y2 * y2, y5 = y4 * y, y6 = y2 * y4;
        const tf2 = tf * tf, tf4 = tf2 * tf2;
        const nf3 = nf * nf * nf, nf5 = nf3 * nf * nf;
        const L = y / (nf * cbf) - ((1 + 2 * tf2 + qf) * y3) / (6 * nf3 * cbf) + ((5 + 28 * tf2 + 24 * tf4 + 6 * qf + 8 * qf * tf2) * y5) / (120 * nf5 * cbf);
        const B = bf - (tf * y2) / (2 * mf * nf) + (tf * (5 + 3 * tf2 + qf - 9 * qf * tf2) * y4) / (24 * mf * nf3) + (tf * (61 + 90 * tf2 + 45 * tf4) * y6) / (720 * mf * nf5);
        return {lng: l0 + L * 180 / PI, lat: B * 180 / PI}
    }

}

const ellipsoids = {
    'CGCS2000': new Ellipsoid('CGCS2000', {
        a: 6378137.0,
        b: 6356752.31414,
        f: 0.003352810681182319,
        e: 0.0818191910428,
        "e'": 0.0820944381519,
        x0: 500000
    }),
    'bj54': new Ellipsoid('bj54', {
        a: 6378245,
        b: 6356863.0188,
        f: 1 / 298.3,
        x0: 500000
    }),
    'xa80': new Ellipsoid('xa80', {
        a: 6378140,
        b: 6356755.2882,
        f: 1 / 298.257,
        e2: 0.00669438499959,
        "e'2": 0.00673950181947,
        x0: 500000
    })
}
